#include "autoptr.hh"

template<typename T>
class dyni : public ptrable
{
public:
    virtual ~dyni() { }
    virtual T value() const = 0;
    virtual void debug() const = 0;
};

template<typename T>
class dyn
{
protected:
    autoptr<dyni<T> > v;
    long pp;
public:
    dyn(autoptr<dyni<T> > p = NULL) : v(p), pp(0) { }
    dyn(const dyn &p) : v(p.v), pp(p.pp) { }
    virtual ~dyn()
    {
    }
    virtual T value() const
    {
        if(!v)
        {
            cerr << "Exception: Tried to get a value of UNSET variable\n";
            throw;
            return T();
        }
        return v->value() + pp;
    }
    virtual void debug() const
    {
        if(!v)
        {
            cerr << "Exception: Tried to get a value of UNSET variable\n";
            throw;
        }
        v->debug();
        if(pp != 0)
        {
            if(pp > 0)cout << '+';
            cout << pp;
        }
    }
    dyn &operator= (const dyn &p)
    {
        v = p.v;
        return *this;
    }
    dyn &operator= (autoptr<dyni<T> > &p)
    {
        v = p;
        return *this;
    }
    dyn &operator= (dyni<T> *p)
    {
        v = p;
        return *this;
    }
    
    dyn &operator= (const T value); // constant
    dyn operator++ () { ++pp; return *this; }
    dyn operator-- () { --pp; return *this; }
    dyn operator++ (int) { dyn tmp = *this; ++pp; return tmp; }
    dyn operator-- (int) { dyn tmp = *this; --pp; return tmp; }
    
    operator T () const
    {
        return value();
    }
};

template<typename T>
class dyntmp : public dyn<T>
{
public:
    dyntmp(const dyn<T> &p) : dyn<T> (p) { }
    /* Temporary dyn<> */
};

template<typename T>
class dynvalue : public dyni<T>
{
    T v;
public:
    dynvalue(T c) : v(c) { }
    virtual T value() const
    {
        return v;
    }
    virtual void debug() const { cout << v; }
};
template<typename T>
class dyn2 : public dyni<T>
{
protected:
    const dyn<T> &L, &R;
public:
    dyn2(const dyn<T> &l, const dyn<T> &r) : L(l), R(r) { }
};

#define dynop(name, op, setop, opchar) \
template<typename T> class dyn##name : public dyn2<T> { public: \
  dyn##name(const dyn<T> &l, const dyn<T> &r) : dyn2<T> (l,r) { } \
  virtual T value() const { return L.value() op R.value(); } \
  virtual void debug() const { cout << '('; L.debug(); cout << opchar; R.debug(); cout << ')'; } \
}; \
template<typename T> \
dyntmp<T> operator op (const dyn<T> &l, const dyn<T> &r) \
{ \
    return dyn<T> (new dyn##name<T> (l, r)); \
} \
template<typename T> \
dyntmp<T> operator setop (dyn<T> &p, const dyn<T> &v) \
{ \
    dyn<T> *q = new dyn<T>(p); \
    return p = new dyn##name<T> (*q, v); \
}

dynop(plus,  +, +=, '+')
dynop(minus, -, -=, '-')
dynop(times, *, *=, '*')
dynop(div,   /, /=, '/')
dynop(mod,   %, %=, '%')

template<typename T>
dyn<T> &dyn<T>::operator= (T value)
{
    return *this = new dynvalue<T> (value);
}
#undef dynop
